//**************************************************************************//
//* FRIDA: flexible rapid interactive data analysis                        *//
//* mystd.cpp: feeble attempts to cure the many deficiencies of C++        *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://www.messen-und-deuten.de/frida                                  *//
//**************************************************************************//

#include <stdlib.h> // otherwise buggy compiler error from gsl
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <glob.h>
#include <wordexp.h>
#include <iostream>
#include <boost/format.hpp>

#include "mystd.h"

using boost::format;

//! declare auxiliary functions.

namespace mystd_aux {
    int tim2dbl(const string inp, double *val);
};


//**************************************************************************//
//*  convert to string                                                     *//
//**************************************************************************//

string strg( const string& val )
{
    return val;
}

string strg( bool val) {
    return val ? "y" : "n";
}

string strg( double val )
{
    return str( format( "%g" ) % val );
}

string strg( int val )
{
    return str( format( "%d" ) % val );
}

string strg( uint val )
{
    return str( format( "%u" ) % val );
}

string strg( unsigned long val)
{
    return str( format( "%lu" ) % val );
}

string strg( char val )
{
    return string(1, val);
}


//**************************************************************************//
//*  yaml                                                                  *//
//**************************************************************************//

//! refine string for yaml output.

string yaml( const string s )
// string -> yaml-output-string
{
    if ( s=="" )
        return "\"\"";
    string out = s;
    uint i = 0;
    // escape "
    while( i < out.size() ){
        if( ( out[i]=='"' || out[i]=='\'' ) && (i==0 || out[i-1]!='\\') )
            out.insert( i, "\\" );
        ++i;
    }
    // if there are special characters, put "" around
    if( s.find( '"' )!=string::npos ||
        s.find( '\'' )!=string::npos ||
        s.find( '\\' )!=string::npos ||
        s.find( ':' )!=string::npos ||
        s.find( '*' )!=string::npos ||
        s.find( '&' )!=string::npos ||
        s.find( '#' )!=string::npos ||
        s.find( '>' )!=string::npos ||
        s.find( '<' )!=string::npos ||
        s.find( '[' )!=string::npos ||
        s.find( ']' )!=string::npos ||
        s.find( '{' )!=string::npos ||
        s.find( '}' )!=string::npos
        )
        out = '"' + out + '"';
    return out;
}


//**************************************************************************//
//*  convert from string                                                   *//
//**************************************************************************//


// ------ should be converted to void, throwing errors -------


bool mystd::str2vec( string inp, vector<double> *V, uint nmax, bool force )
{
    using namespace mystd_aux;
    using namespace std;

    // cout << "DEBUG mystd [" << inp << "]\n";
    V->clear();

    while(inp.size() && isspace(inp[0])) inp.erase(inp.begin());
    // while(inp.size() && strchr(" \t.\n",inp[inp.size()-1])) 
    //    inp.erase(inp.end()-1);

    // cout << "DEBUG  core [" << inp << "]\n";
    if (inp==string("")) {
        // cout << "DEBUG  empty\n";
        return true;
    }
    // cout << "DEBUG  nonempty\n";

    string::size_type ji=0, jf;
    double val;
    char dummy;
    while (ji<inp.size()) {
        if ((jf=inp.find_first_of(", \t\n", ji)) == string::npos) 
            jf = inp.size();
        // cout << " look until " << jf << "\n";
        if (sscanf(inp.substr(ji,jf-ji).c_str(), "%lg%c", &val, &dummy)
            ==1) {
            V->push_back(val);
        } else if ( force ) {
            cout << "! str2vec: invalid value ["
                 << inp.substr(ji,jf-ji) << "] at pos " << ji
                 << " in [" << inp << "]\n";
            return false;
        }
        if (nmax && V->size()>=nmax) break; // do not decode the tail
        // cout << " found "<<ji<<":"<<jf<<" value "<<val<<"\n";
        ji = inp.find_first_not_of(", \t\n", jf);
        // cout << " continue at " << ji << "\n";
    }

    return true;
}

bool mystd::any2dbl( const string& inp, double *val )
{
    if(!inp.size()) {
        cout << "expecting double, found empty input\n";
        return false;
    }
    if( inp=="-inf" || inp=="-oo" ){
        *val = -INFINITY;
        return true;
    }
    if( inp=="inf" || inp=="oo" || inp=="+inf" || inp=="+oo" ){
        *val = INFINITY;
        return true;
    }
    if(!strchr("+-0123456789.", inp[0])) {
        cout << "expecting double, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    double mant; int expt;
    if (sscanf(inp.c_str(), "%lgm%d", &mant, &expt)==2) {
        *val = mant * pow(10.e0, (double)(-expt));
        return true;
    }
    if (sscanf(inp.c_str(), "%lg", val)==1)
        return true;
    cout << "expecting double, sscanf failed\n";
    return false;
}

bool mystd::any2int( const string& inp, int *val )
{
    if(!inp.size()) {
        cout << "expecting integer, found empty input\n";
        return false;
    }
    if(!strchr("+-0123456789", inp[0])) {
        cout << "expecting integer, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    if (sscanf(inp.c_str(), "%d", val)==1) return true;
    cout << "expecting integer, sscanf failed\n";
    return false;
}

bool mystd::any2uint( const string& inp, uint *val )
{
    if(!inp.size()) {
        cout << "expecting integer, found empty input\n";
        return false;
    }
    if(!strchr("0123456789", inp[0])) {
        cout << "expecting integer, found bad char[0] in [" 
             << inp << "]\n";
        return false;
    }
    if (sscanf(inp.c_str(), "%u", val)==1) return true;
    cout << "expecting integer, sscanf failed\n";
    return false;
}

//**************************************************************************//
//*  string operations                                                     *//
//**************************************************************************//

string mystd::strip(
    string const& str, ///< the original string
    char const* sepSet ///< C string with characters to be dropped
)
{
    string::size_type const first = str.find_first_not_of(sepSet);
    return ( first==std::string::npos )
        ? std::string()
        : str.substr(first, str.find_last_not_of(sepSet)-first+1);
}

void mystd::string_extract_word( const string& in, string *out1, string *out2 )
{
    // one word (*out1) is extracted; leading " \t" are retained.
    // in the tail (*out2) leading " \t" are deleted.
    // "\n" starts always a new word.
    // leading "`" makes *out1 go to next "`" or to end of line.

    string::size_type j0=0, j1, j2, jf=in.size();

    // handle special case "`"
    if (in[0]=='`') { 
        j1 = in.find_first_of('`', 1);
        if (j1==string::npos) {
            *out1 = in.substr(1);
            *out2 = "";
        } else {
            *out1 = in.substr(1,j1-1);
            j2 = in.find_first_not_of(" \t", j1+1);
            if (j2!=string::npos) 
                *out2 = in.substr(j2);
            else
                *out2 = "";
        }
        return;
    }

    while (j0<jf && strchr(" \t",in[j0])) j0++;
    j1=j0+1;
    while (j1<jf && !strchr(" \t\n",in[j1])) j1++;
    j2=j1;
    while (j2<jf && strchr(" \t",in[j2])) j2++;
    *out1 = in.substr(0,j1);
    if (j2<jf) 
        *out2 = in.substr(j2);
    else
        *out2 = string("");
    //cout << "from ["<<in<<"] word ["<<*out1<<"] and tail ["<<*out2<<"]\n";
}

void mystd::string_extract_line( const string& in, string *out1, string *out2 )
{
    string line;
    string::size_type i = in.find('\n');
    if (i==string::npos) {
        *out1 = in;
        *out2 = string("");
    } else {
        *out1 = in.substr(0,i);
        *out2 = in.substr(i+1);
    }
    //cout << "from ["<<in<<"] line ["<<*out1<<"] and tail ["<<*out2<<"]\n";
}


//! Split string at whitespaces, set vector of words.

void mystd::string2words( const string& in, vector<string>& out )
{
    const string delimiters = " \t";
    // skip delimiters at beginning:
    string::size_type lastPos = in.find_first_not_of( delimiters, 0 );
    // find first non-delimiter:
    string::size_type pos = in.find_first_of( delimiters, lastPos );
    while ( pos!=string::npos || lastPos!=string::npos ) {
        // found a token, add it to the vector:
        out.push_back( in.substr(lastPos, pos-lastPos) );
        // skip delimiters:
        lastPos = in.find_first_not_of( delimiters, pos );
        // find next non-delimiter:
        pos = in.find_first_of( delimiters, lastPos );
    }
}


//**************************************************************************//
//*  improved libc calls                                                   *//
//**************************************************************************//

void mystd::system( string cmd )
{
    int ret = ::system( cmd.c_str() );
    if( ret )
        throw "Command '" + cmd + "' returned error code " + strg(ret);
}

//**************************************************************************//
//*  time                                                                  *//
//**************************************************************************//

//! Seconds since the epoch. Return them as double (WHY??)
//  only used for calls to myasctime => OBSOLETE

double mystd::justnow()
{
    struct timeval exactlynow;
    gettimeofday(&exactlynow, NULL);
    return (double) exactlynow.tv_sec;
}

//! Return time as string in ctime/asctime standard format.
//  OBSOLETE: better use strftime

string mystd::myasctime( double time, int offs )
{
    long timin = (long) time;
    char buf[40];
    sprintf(buf, "%.24s", asctime(localtime(&timin)));
    return string(buf+offs);
}


//**************************************************************************//
//*  rounding                                                              *//
//**************************************************************************//

//! Round given value to given number of digits.

double mystd::round_decimal( double val, double digits )
{
    if (val==0 || digits<=0) return 0;
    double v = fabs(val);
    int n = (int) floor ( digits - log10(v));
    // cout << " n=" << n;
    double p = pow(10., n);
    double r = round(v*p) / p;
    return (val<0 ? -1 : +1) * r;
}


//**************************************************************************//
//*  vector operations                                                     *//
//**************************************************************************//

//! insert val into vector *V, presuming the latter is sorted

void mystd::merge( vector<double> *V, double val )
{
    for (uint i=0; i<V->size(); ++i) {
        if (val>(*V)[i]) continue;
        if (val==(*V)[i]) return;
        V->insert(V->begin()+i, val);
        return;
    }
    V->push_back(val);
}

/*
bool mystd::get_const(vector<double> *V, double *val)
    // check whether *V has a constant value; write this value to *val
{
    double v;
    if (V->size()==0) return false; // empty vector
    v = (*V)[0];
    for (uint i=1; i<V->size(); ++i)
        if ((*V)[i]!=v) return false; // vector not constant
    if (val) *val = v;
    return true;
}
*/

void mystd::unique( vector<double> *V, double tolabs, double tolrel )
{
    // tolrel z.Zt. nicht benutzt
    vector<double> aux;
    double diff;
    if (!V->size()) return;
    aux.push_back((*V)[0]);
    for (uint ii=1; ii<V->size(); ++ii)
        if((diff=fabs((*V)[ii]-aux.back()))>tolabs)
            aux.push_back((*V)[ii]);
    *V = aux;
}

//! detect sorting (strictly de- or increasing: -1 or1; else: 0).

int mystd::sorted( const vector<double>& V )
{
    uint n = V.size();
    if( n < 2 )
        throw string( "vector too short to determine sorting" );
    int ret;
    if      ( V[n-1] > V[0] )
        ret = 1;
    else if ( V[n-1] < V[0] )
        ret = -1;
    else
        return 0;
    for( uint i=2; i<n; ++i )
        if ( ret*(V[i]-V[i-1]) <= 0 )
            return 0;
    return ret;
}

//! detect equidistant grid, set *step.

bool mystd::is_equidist( double *step, const vector<double>& V )
{
    uint n = V.size();
    if( n < 2 )
        return false; // vector too short to determine equidistance
    *step = ( V[n-1] - V[0] ) / (n-1);
    for( uint i=1; i<n-1; ++i )
        if( fabs( V[i] - (V[0]+i*(*step)) ) > 1e-12*fabs(*step)+1e-200 )
            return false;
    return true;
}

//! sum over vector elements.

double mystd::sum( const vector<double>& V )
{
    double ret = 0;
    for( uint i=0; i<V.size(); ++i )
        ret += V[i];
    return ret;
}

//**************************************************************************//
//*  file names, globbing, file read                                       *//
//**************************************************************************//


//! Test whether file exists.

bool mystd::file_exists( const string& fname )
{
    struct stat fileInfo;
    int iStat = stat( fname.c_str(), &fileInfo );
    if ( iStat == 0 ) // we got attributes, so the file obviously exists
        return true;
    else
        return false; // either file doesn't exist or we lack permissions
}


//! Analyse file name: divide "dir/short.ext" into "dir", "short", "ext".

void mystd::fname_divide( const string& fname,
                          string *fdir, string *fshort, string *fext )
// if some of the three return values are not needed, the user may supply 0
// instead of a pointer to a string.
{
    uint idir = fname.rfind("/") + 1;
    if( fdir ) *fdir = fname.substr( 0, idir );

    uint iext = fname.rfind(string(".")) + 1;
    if( iext<=idir ){
        if( fext ) *fext = "";
        if( fshort ) *fshort = fname.substr( idir );
    } else {
        if( fext ) *fext = fname.substr( iext );
        if( fshort ) *fshort = fname.substr( idir, iext-idir-1 );
    }
}


//! Bash expansion of file name.

string mystd::wordexp_unique( const string& s )
{
    wordexp_t p;
    if( wordexp( s.c_str(), &p, 0 ) )
        throw "cannot expand " + s;
    if( p.we_wordc!=1 )
        throw "expansion of " + s + " not unique";
    string ret = p.we_wordv[0];
    wordfree( &p );
    return ret;
}


//! Expansion of file list.

void mystd::glob_file_list(
    const string& pattern, const string& extension, vector<string> *fnames )
{
    // add extension to input pattern:
    string extended_pattern = "";
    vector<string> words;
    string2words( pattern, words );
    if ( !words.size() )
        throw string( "empty file list" );
    for ( int i=0; ; ) {
        string fmain, fext;
        fname_divide( words[i], 0, &fmain, &fext );
        extended_pattern += words[i];
        if ( fext=="" && extension!="" )
            extended_pattern += "." + extension;
        if ( ++i>=words.size() )
            break;
        extended_pattern += " ";
    }

    wordexp_t p;
    if( wordexp( extended_pattern.c_str(), &p, 0 ) )
        throw "cannot expand " + extended_pattern;
    if( p.we_wordc==0 )
        throw "expansion of " + extended_pattern + " is empty";
    for ( int i=0; i<p.we_wordc; ++i )
        fnames->push_back( p.we_wordv[i] );
    wordfree( &p );
}


//! Read one line from file. 

// OBSOLETE: replace by simpler version from read-plus

int mystd::freadln( FILE *fd, string *s )
{
    char c = (char) fgetc(fd);
    if( c==EOF )
        return -1;
    if( c=='\n' ){
        *s = "";
        return 0;
    } else if ( c=='\r' ) {
        if( fgetc(fd)=='\n' ){
            *s = "";
            return 0;
        } else
            throw string( "CR not followed by LF" );
    }
    *s = c;
    while ( 1 ) {
        c = (char) fgetc(fd);
        if ( c=='\n' || c==EOF )
            return s->size();
        else if ( c=='\r' ){
            if( fgetc(fd)=='\n' )
                return s->size();
            else
                throw string( "CR not followed by LF" );
        }
        *s += c;
    }
}


//! Construct tmp file name that is not currently in use.

string mystd::next_tmp_file( const string& path_format )
{
    const int tmpmax=1024;
    for( int i=0; i<tmpmax; ++i ){
        string fname = str( format( path_format ) % i );
        if ( !mystd::file_exists( fname ) )
            return fname;
        if( i==tmpmax/2 )
            cerr << "WARNING: there more than " << i <<
                " temporay files of form " << path_format <<
                "; please clean up soon\n";
    }
    throw "Too many temporary files in use";
}

//**************************************************************************//
//*  GSL integration                                                       *//
//**************************************************************************//

#include <gsl/gsl_integration.h>

double mystd::Integrate( double (*func_) (double, void*), void *data_, 
                         double low, double hig,
                         uint mode, double epsabs,
                         double epsrel )
{
    gsl_function F;
    F.function = func_;
    F.params = data_;
#define nwork 20000
    gsl_integration_workspace *work = gsl_integration_workspace_alloc( nwork );
    double val, err;
    size_t neval;

    if      ( isinf(low)==-1 && isinf(hig)==+1 )
        gsl_integration_qagi( &F, 
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( isinf(low)==-1 )
        gsl_integration_qagil( &F, hig,
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( isinf(hig)==+1 )
        gsl_integration_qagiu( &F, low,
                             epsabs, epsrel, nwork,
                             work, &val, &err );
    else if ( mode==0 )
        // QNG non-adaptive Gauss-Kronrod integration
        gsl_integration_qng( &F, low, hig,
                             epsabs, epsrel,
                             &val, &err, &neval );
    else if ( mode>0 && mode<7 ){
        int rule;
        switch( mode ){
        case 1: rule = GSL_INTEG_GAUSS61; break;
        case 2: rule = GSL_INTEG_GAUSS51; break;
        case 3: rule = GSL_INTEG_GAUSS41; break;
        case 4: rule = GSL_INTEG_GAUSS31; break;
        case 5: rule = GSL_INTEG_GAUSS21; break;
        case 6: rule = GSL_INTEG_GAUSS15; break;
        }
        // QAG adaptive integration
        gsl_integration_qag( &F, low, hig,
                             epsabs, epsrel, nwork,
                             rule, work, &val, &err );
    }
    else if ( mode==7 )
        // QAGS adaptive integration with singularities
        gsl_integration_qags( &F, low, hig,
                              epsabs, epsrel, nwork,
                              work, &val, &err );
    else
        throw string( "invalid mode in mystd::Integrate" );

    gsl_integration_workspace_free( work );
    return val;
}
